#include "capture.h"

#include "common.h"
#include "log.h"

#include <pcap.h>
// malloc()
#include <stdlib.h>
// strdup()
#include <string.h>
// constants, struct ether_header
#include <net/ethernet.h>

// object of change - to fit only ARP/ND needs (without VLAN)
// ARP for IPv4 = 28
// IPv6 header = 40; ICMPv6 header = 4
#define SNAPLEN			ETHER_HDR_LEN + 40 + 4

// not included in pcap.h prior to 1.1.0
#ifndef PCAP_NETMASK_UNKNOWN
#define PCAP_NETMASK_UNKNOWN	0xffffffff
#endif

int is_bcast(pcap_if_t *dev)
{
	pcap_addr_t *addr;
	for (addr = dev->addresses; addr != NULL; addr = addr->next)
		if (addr->broadaddr != NULL) return 1;
	return 0;
}

int ifname_is_bcast_dev(char *ifname, pcap_if_t *alldevs)
{
	pcap_if_t *dev;
	for (dev = alldevs; dev != NULL; dev = dev->next)
		if (strcmp(dev->name, ifname) == 0)
			if (is_bcast(dev))
				return 1;
	return 0;
}

int ifname_is_capturing(char *ifname, struct capture *captures)
{
	struct capture *cap;
	for (cap = captures; cap != NULL; cap = cap->next) {
		if (strcmp(ifname, cap->ifname) == 0) {
			//printf("already capturing dev %s\n", ifname);
			return 1;
		}
	}
	//printf("dev %s not found in capture list\n", ifname);
	return 0;
}

int set_filter(pcap_t *pcap_handle)
{
	struct bpf_program bpf;
	if (pcap_compile(pcap_handle, &bpf, FILTER, 0, PCAP_NETMASK_UNKNOWN)
			== 0) {
		if (pcap_setfilter(pcap_handle, &bpf) == 0) {
			// cleanup after successful install 
			pcap_freecode(&bpf);
			return 1;
		} else {
			log_print_time();
			fprintf(logfile, "pcap_setfilter failed\n");
		}
		pcap_freecode(&bpf);
	} else {
		log_print_time();
		fprintf(logfile, "pcap_compile failed\n");
	}
	return 0;
}

pcap_t *start_capture(char *ifname)
{
	pcap_t *handle;
	char errbuf[PCAP_ERRBUF_SIZE];
	// no promisc (try) and no TO
	handle = pcap_open_live(ifname, SNAPLEN, 0, 0, errbuf);
	if (handle != NULL) {
		if (pcap_datalink(handle) == DLT_EN10MB) {
			if (pcap_setnonblock(handle, 1, errbuf) == 0) {
				// #ifdef pcap_version >= ... set_direction(Rx)
				// to reduce load
				// (CentOS 4 has older libpcap)
				if (set_filter(handle)) {
					return handle;
				}
			} else {
				log_print_time();
				fprintf(logfile, "pcap_setnonblock failed\n");
				fprintf(logfile, "%s\n", errbuf);
			}
		} else {
			log_print_time();
			fprintf(logfile, "pcap_datalink returned non Ethernet\n");
		}
		pcap_close(handle);
	} else {
		log_print_time();
		fprintf(logfile, "pcap_open_live failed\n");
		fprintf(logfile, "%s\n", errbuf);
	}
	return NULL;
}

struct capture *init_capture(char *ifname)
{
	struct capture *newcap;
	newcap = malloc(sizeof(struct capture));
	if (newcap != NULL) {
		newcap->ifname = strdup(ifname);
		if (newcap->ifname != NULL) {
			newcap->pcap = start_capture(newcap->ifname);
			if (newcap->pcap != NULL) {
				log_print_time();
				fprintf(logfile, "nonblocking capture started on %s\n", ifname);
				return newcap;
			}
			free(newcap->ifname);
		} else {
			log_print_time();
			fprintf(logfile, "strdup failed\n");
		}
		free(newcap);
	} else {
		log_print_time();
		fprintf(logfile, "malloc failed\n");
	}
	return NULL;
}

// close capture, free memory
void free_capture(struct capture *cap)
{
	log_print_time();
	fprintf(logfile, "closing capture on %s\n", cap->ifname);
	pcap_close(cap->pcap);
	free(cap->ifname);
	free(cap);
}

void upkeep_captures(struct capture **captures)
{
	char errbuf[PCAP_ERRBUF_SIZE];
	pcap_if_t *alldevs, *dev;
	struct capture *cap, *prevcap, *nextcap, *newcap;
	//print_bcast_devs();
	if (pcap_findalldevs(&alldevs, errbuf) == -1) {
		log_print_time();
		fprintf(logfile, "pcap_findalldevs failed\n");
		fprintf(logfile, "%s\n", errbuf);
		// if findalldevs failed, we can't upkeep captures
		return;
	}
	if (alldevs == NULL) {
		log_print_time();
		fprintf(logfile, "No devices found, check privileges.\n");
	}
	// stop capturing on removed interfaces
	for (cap = *captures; cap != NULL;) {
		nextcap = cap->next;
		if (! ifname_is_bcast_dev(cap->ifname, alldevs)) {
			// remove cap from list
			if (cap == *captures)
				// cut head
				*captures = nextcap;
			else
				// or link previous to next
				prevcap->next = nextcap;
			free_capture(cap);
		} else
			// store non removed list member for future linking
			prevcap = cap;
		cap = nextcap;
	}
	// start capturing on new broadcast interfaces
	for (dev = alldevs; dev != NULL; dev = dev->next) {
		if (is_bcast(dev) &&
				! ifname_is_capturing(dev->name, *captures)) {
			//printf("broadcast device: %s\n", dev->name);
			newcap = init_capture(dev->name);
			if (newcap != NULL) {
				// add newcap to head
				newcap->next = *captures;
				*captures = newcap;
			} else {
				//printf("new capture failed\n");
			}
		}
	}	
	pcap_freealldevs(alldevs);
}

void free_captures(struct capture *captures)
{
	struct capture *cap, *nextcap;
	for (cap = captures; cap != NULL;) {
		nextcap = cap->next;
		free_capture(cap);
		cap = nextcap;
	}
}
