#include "db.h"

#include "common.h"
#include "confile.h"
#include "list.h"
#include "log.h"

#include <mysql/mysql.h>

// printf()
#include <stdio.h>
// str*()
#include <string.h>
// inet_ntop()
#include <arpa/inet.h>

// global for this .c
static MYSQL *handler = NULL;

void db_error_log(void)
{
	log_print_time();
	fprintf(logfile, "mysql_errno: %u\n", mysql_errno(handler));
	log_print_time();
	fprintf(logfile, "mysql_error: %s\n", mysql_error(handler));
}

void db_close(void)
{
	if (handler != NULL) {
		log_print_time();
		fputs("calling myslq_close()\n", logfile);
		mysql_close(handler);
		// to the beginning
		handler = NULL;
	}
}

int db_connect(struct conf *conf)
{
	if (handler == NULL) {
		//fputs("handler is NULL, calling mysql_init()\n", logfile);
		if ((handler = mysql_init(NULL)) == NULL) {
			log_print_time();
			fputs("mysql_init() failed\n", logfile);
			return 0;
		}
	}
	// calling mysql_ping() here leads to segfault (known error)
	// and reconnect (and thread_id() and...) behavior depends on versions
	// so avoid this whole thing and close in case of error
	if (mysql_real_connect(handler, conf->dbhost, conf->dbuser,
				conf->dbpass, conf->dbname, 0, NULL, 0)
			== NULL) {
		log_print_time();
		fputs("mysql_real_connect() failed\n", logfile);
		db_error_log();
		// cleanup
		db_close();
		return 0;
	}
	return 1;
}

void sprint_full_mac(char *string, const u_char *mac)
{
	int i;
	sprintf(string, "%02x", mac[0]);
	for (i = 1; i < 6; i++)
		sprintf(string + i * 3 - 1, ":%02x", mac[i]);
}

int db_query(const char *query)
{
	if (mysql_query(handler, query)) {
		log_print_time();
		fprintf(logfile, "mysql_query(%s) failed\n", query);
		db_error_log();
		// cleanup
		db_close();
		return 0;
	}
	return 1;
}

int db_insert_arp(struct arp_row *arp_row)
{
	char q_begin[] = "insert arp values (\"";
	char q_mac[] = "xx:xx:xx:xx:xx:xx";
	char q_sep1[] = "\", \"";
	char q_ip4[INET_ADDRSTRLEN];
	char q_sep2[] = "\", ";
	char q_dup[] = ") on duplicate key update time = ";
	// D. A. constant means 2x ulong (64b, 20B printed)
	// + 2x (im)possible minus sign
	// sizeof() will be better
	char query[strlen(q_begin) + strlen(q_mac) + strlen(q_sep1)
	+ INET_ADDRSTRLEN + strlen(q_sep2) + strlen(q_dup) + 42];
	if (handler == NULL) {
		log_print_time();
		fputs("handler is NULL, exiting\n", logfile);
		return 0;
	}
	sprint_full_mac(q_mac, arp_row->mac);
	inet_ntop(AF_INET, arp_row->ip4, q_ip4,	INET_ADDRSTRLEN);
	sprintf(query, "%s%s%s%s%s%ld%s%ld", q_begin, q_mac, q_sep1, q_ip4,
			q_sep2, arp_row->time, q_dup, arp_row->time);
	return db_query(query);
}

int db_flush_arp(struct list *arp)
{
	while (arp->head != NULL)
		if (db_insert_arp(list_get_head(arp)))
			list_delete_head(arp);
		else
			return 0;
	return 1;
}

void sprint_full_ip6(char *string, const u_char *ip6)
{
	int i;
	sprintf(string, "%02x%02x", ip6[0], ip6[1]);
	for (i = 1; i < 8; i++)
		sprintf(string + i * 5 - 1, ":%02x%02x",
				ip6[2 * i], ip6[2 * i + 1]);
}

int db_insert_nd(struct nd_row *nd_row)
{
	char q_begin[] = "insert nd values (\"";
	char q_mac[] = "xx:xx:xx:xx:xx:xx";
	char q_sep1[] = "\", \"";
	char q_ip6[INET6_ADDRSTRLEN];
	char q_sep2[] = "\", ";
	char q_dup[] = ") on duplicate key update time = ";
	// D. A. constant means 2x ulong (64b, 20B printed)
	// + 2x (im)possible minus sign
	char query[strlen(q_begin) + strlen(q_mac) + strlen(q_sep1)
	+ INET6_ADDRSTRLEN + strlen(q_sep2) + strlen(q_dup) + 42];
	if (handler == NULL) {
		log_print_time();
		fputs("handler is NULL, exiting\n", logfile);
		return 0;
	}
	sprint_full_mac(q_mac, nd_row->mac);
	sprint_full_ip6(q_ip6, nd_row->ip6);
	sprintf(query, "%s%s%s%s%s%ld%s%ld", q_begin, q_mac, q_sep1, q_ip6,
			q_sep2, nd_row->time, q_dup, nd_row->time);
/*	fprintf(logfile, "q: %s, il: %ld, fl: %ld\n", query,
			strlen(q_begin) + strlen(q_mac) + strlen(q_sep1)
			+ INET6_ADDRSTRLEN + strlen(q_sep2) + strlen(q_dup)
			+ 42, strlen(query));
*/	return db_query(query);
}

int db_flush_nd(struct list *nd)
{
	while (nd->head != NULL)
		if (db_insert_nd(list_get_head(nd)))
			list_delete_head(nd);
		else
			return 0;
	return 1;
}

int db_flush_queue(struct queue *queue)
{
	if (! db_flush_arp(queue->arp))
		return 0;
	if (! db_flush_nd(queue->nd))
		return 0;
	return 1;
}
