#include "packet.h"

// how often do the packet processing (usec)
#define PACKET_SCAN_PERIOD	1000000

// struct capture
#include "common.h"
// struct list
#include "list.h"
// FILTER
#include "capture.h"
#include "log.h"

#include <pcap.h>
// constants, struct ether_header
#include <net/ethernet.h>
// ntohs()
#include <arpa/inet.h>
// arphdr
#include <net/if_arp.h>
// memcmp()
#include <string.h>
// sleep()
#include <unistd.h>
// gettimeofday()
#include <time.h>
// malloc()
#include <stdlib.h>

// what's missing in older ethernet.h
#ifndef ETHERTYPE_VLAN
#define	ETHERTYPE_VLAN		0x8100
#endif
#ifndef ETHERTYPE_IPV6
#define	ETHERTYPE_IPV6		0x86dd
#endif

int arp_ipv4_header(struct arphdr *header, time_t time)
{
	unsigned short int htype, ptype, oper;
	htype = ntohs(header->ar_hrd);
	if (htype != ARPHRD_ETHER) {
		log_print_time();
		fprintf(logfile, "unexpected ARP HTYPE: 0x%04x (arrived at %ld)\n", htype, time);
		return 0;
	}
	ptype = ntohs(header->ar_pro);
	if (ptype != ETHERTYPE_IP) {
		log_print_time();
		fprintf(logfile, "unsupported ARP PTYPE: 0x%04x (arrived at %ld)\n", ptype, time);
		return 0;
	}
	if (header->ar_hln != ETHER_ADDR_LEN) {
		log_print_time();
		fprintf(logfile, "unexpected ARP HLEN: 0x%02x (arrived at %ld)\n", header->ar_hln, time);
		return 0;
	}
	// IPv4 length
	if (header->ar_pln != 4) {
		log_print_time();
		fprintf(logfile, "unexpected ARP PLEN: 0x%02x (arrived at %ld)\n", header->ar_pln, time);
		return 0;
	}
	oper = ntohs(header->ar_op);
	// process only this
	if (oper == ARPOP_REQUEST || oper == ARPOP_REPLY) return 1;
	// pass others
	return 0;
}

void process_arp(time_t time, const u_char *data, struct list *arp)
{
	const u_char *ethsrc, *arpsha, *arpspa;
	struct arp_row *arp_data;
	if (!arp_ipv4_header((struct arphdr *)(data + ETHER_HDR_LEN), time))
		return;
	ethsrc = data + ETHER_ADDR_LEN;
	arpsha = data + ETHER_HDR_LEN + sizeof(struct arphdr);
	// compare Ethernet src MAC and ARP src MAC for sake of completeness
	if (memcmp(ethsrc, arpsha, ETHER_ADDR_LEN) != 0) {
		log_print_time();
		fprintf(logfile, "Ethernet and ARP src MAC differ (arrived at %ld)\n", time);
		return;
	}
	arpspa = arpsha + ETHER_ADDR_LEN;
	// avoid ARP probe (SPA is 0.0.0.0)
	// no need for ntohl here - unsigned is enough for zero
	if (*(uint32_t *)arpspa == 0) return;
	arp_data = malloc(sizeof(struct arp_row));
	if (arp_data == NULL) return;
	arp_data->time = time;
	// copy MAC address and IPv4 address together
	memcpy(arp_data->mac, arpsha, ETHER_ADDR_LEN + 4);
	list_append(arp, arp_data);
}

void process_ipv6(time_t time, const u_char *data, struct list *nd)
{
	const u_char *ethsrc, *ip6src, *nh, *type;
	const u_char zero_ip6a[16] = {0};
	struct nd_row *nd_data;
	// Next Header is 58 for ICMPv6
	nh = data + ETHER_HDR_LEN + 6;
	if (*(uint8_t *)nh != 58) {
		log_print_time();
		fprintf(logfile, "unexpected IPv6 Next Header: %d (arrived at %ld)\n", *(uint8_t *)nh, time);
		return;
	}
	// ICMPv6 Type is the first byte of IPv6 packet payload
	type = data + ETHER_HDR_LEN + 40;
	// process only Neighbor Solicitation or Advertisement
	if (!(*(uint8_t *)type == 135 || *(uint8_t *)type == 136)) return;
	ip6src = data + ETHER_HDR_LEN + 8;
	// pass ::
	if (memcmp(ip6src, zero_ip6a, 16) == 0) return;
	ethsrc = data + ETHER_ADDR_LEN;
	nd_data = malloc(sizeof(struct nd_row));
	if (nd_data == NULL) return;
	nd_data->time = time;
	// MAC address
	memcpy(nd_data->mac, ethsrc, ETHER_ADDR_LEN);
	// IPv6 address
	memcpy(nd_data->ip6, ip6src, 16);
	list_append(nd, nd_data);
}

void callback(u_char *queue, const struct pcap_pkthdr *header,
		const u_char *data)
{
	u_int16_t ether_type;
/*	int i;
	printf("packet arrived at %d.%d\n",
			header->ts.tv_sec, header->ts.tv_usec);
	printf("len: %d, caplen: %d", header->len, header->caplen);
	for (i = 0; i < header->caplen; i++) {
		if (!(i % 16)) printf("\n");
		printf("%02x ", data[i]);
	}
	printf("\n");*/
	// 14 is enough for ether_type, but
	// minimum payload to examine is IPv4 ARP (28 bytes)
	if (header->caplen < ETHER_HDR_LEN + 28) {
		log_print_time();
		fprintf(logfile, "captured packet too short (%d bytes, arrived at %ld)\n", header->caplen, header->ts.tv_sec);
		return;
	}
	ether_type = ntohs(((struct ether_header *)data)->ether_type);
	switch (ether_type) {
		case ETHERTYPE_ARP:
			process_arp(header->ts.tv_sec, data,
					((struct queue *)queue)->arp);
			break;
		case ETHERTYPE_VLAN:
			// for now we leave nested packets
			// passed through the filter
			//printf("VLAN\n");
			return;
		case ETHERTYPE_IPV6:
			if (header->caplen < ETHER_HDR_LEN + 44) {
				log_print_time();
				fprintf(logfile, "captured IPv6 packet too short (%d bytes, arrived at %ld)\n", header->caplen, header->ts.tv_sec);
				return;
			}
			process_ipv6(header->ts.tv_sec, data,
					((struct queue *)queue)->nd);
			break;
		default:
			log_print_time();
			fprintf(logfile, "another ether type passed through the filter \"%s\": 0x%04x (arrived at %ld)\n", FILTER, ether_type,
					header->ts.tv_sec);
			return;
	}
}

void process_packets(struct capture *captures, struct queue *queue)
{
	struct capture *cap;
	for (cap = captures; cap != NULL; cap = cap->next)
		// on CentOS 4.9 (pcap 0.8.3) and CentOS 5.10 (pcap 0.9.4)
		// pcap_dispatch returns no more than one packet
		// regardless what man page is saying about -1 (kernel issue?)
		// so there will be always one useless call
		while (pcap_dispatch(cap->pcap, -1, callback, (u_char *)queue)
				> 0);
}

void packet_loop(struct capture *captures, struct queue *queue)
{
	struct timeval start_time, end_time;
	int time_err;
	suseconds_t duration, sleep_time;
	time_err = gettimeofday(&start_time, NULL);
	process_packets(captures, queue);
	time_err += gettimeofday(&end_time, NULL);
	if (time_err) {
		//printf("sleeping one second...\n");
		sleep(1);
	} else {
		duration = 1000000
			* (end_time.tv_sec - start_time.tv_sec)
			+ (end_time.tv_usec - start_time.tv_usec);
		//printf("duration: %d", duration);
		sleep_time = PACKET_SCAN_PERIOD - duration;
		if (sleep_time > 0) {
			//printf(", sleeping %d usec...", sleep_time);
			usleep(sleep_time);
		}
		//printf("\n");
	}
}
